# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/views.ttest.ipynb (unless otherwise specified).

__all__ = ['TTestLandingView', 'TTestLanding', 'IndependentSamplesEstimates', 'IndependentSamples',
           'PairedSamplesEstimates', 'PairedSamples', 'TTestEstimates', 'TTest', 'SingleSample', 'DiffPairedSamples']

# Cell
from ..widgets import FloatText, IntText, Text, Grid, AlignedLabel, HTMLMath, Button, \
    GridBox, BoundedFloatText, RadioButtons
from ..widget_containers import Box, VBox, HBox
from ..mvc import Controller
from .error_bars import ErrorBars

# Cell
import ipywidgets as ipyw
from traitlets import observe

# Cell
class TTestLandingView(VBox):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        style= {'description_width': 'initial'}
        self.label1 = Text(description='Group 1', style=style, layout=ipyw.Layout(width='140px'))
        self.label2 = Text(description='Group 2', style=style, layout=ipyw.Layout(width='140px'))
        self.children = (self.label1, self.label2)

# Cell
class TTestLanding(TTestLandingView, Controller):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.setLanding()
        self.label1.observe(self.observe_label1, 'value')
        self.label2.observe(self.observe_label2, 'value')

    def setLanding(self):
        self.label1.value = self.model.label1
        self.label2.value = self.model.label2

    def observe_label1(self, change):
        self.model.label1 = change['new']
        return change['new']

    def observe_label2(self, change):
        self.model.label2 = change['new']
        return change['new']

# Cell
class IndependentSamplesEstimates(Box):

    def __init__(self, **kwargs):
        lo = {'justify_content': 'space-between', 'flex_flow': 'row wrap'}
        super().__init__(layout=ipyw.Layout(**lo), **kwargs)


        ''' self.grid '''
        style={'description_width': '35px'}
        self.m1 = FloatText(step=0.01, description=r'$$M_1$$', style=style, layout=ipyw.Layout(width='100px'))
        self.m2 = FloatText(step=0.01,description=r'$$M_2$$', style=style, layout=ipyw.Layout(width='100px'))
        self.sd1 = FloatText(step=0.01,description=r'$$SD_1$$', style=style, layout=ipyw.Layout(width='100px'))
        self.sd2 = FloatText(step=0.01,description=r'$$SD_2$$', style=style, layout=ipyw.Layout(width='100px'))
        self.ratio = FloatText(step=0.01,description=r'$$n_1 / n_2$$', style=style, layout=ipyw.Layout(width='100px'))
        self.m1_label = AlignedLabel(value='Mean', alignment='left')
        self.m2_label = AlignedLabel(value='Mean', alignment='left')
        self.sd1_label = AlignedLabel(value='Standard Deviation', alignment='left')
        self.sd2_label = AlignedLabel(value='Standard Deviation', alignment='left')
        self.ratio_label = AlignedLabel(value='Allocation Ratio', alignment='left')

        self.grid = Grid(6, 2, layout=ipyw.Layout(grid_template_columns='112px auto'))
        self.grid.layout.margin = '0 0 25px 0'
        self.grid[0, 0] = AlignedLabel(value='Estimate', alignment='left')
        self.grid[0, 1] = AlignedLabel(value='Parameter', alignment='left')
        self.grid[1,0] = self.m1
        self.grid[2,0] = self.m2
        self.grid[3,0] = self.sd1
        self.grid[4,0] = self.sd2
        self.grid[5,0] = self.ratio
        self.grid[1, 1] = self.m1_label
        self.grid[2, 1] = self.m2_label
        self.grid[3, 1] = self.sd1_label
        self.grid[4, 1] = self.sd2_label
        self.grid[5, 1] = self.ratio_label

        d_label = HTMLMath(value=r'$$\frac{M_1 - M_2}{\sqrt{\frac{ratio*SD_1^2 + SD_2^2}{ratio+1}}}=d=$$', layout=ipyw.Layout(margin='5'))
        self.d = Button(disabled=True, layout=ipyw.Layout(width='65px'))
        box_layout = {'justify_content': 'flex-end'}
        self.d_box = HBox(children=(d_label, self.d), layout=ipyw.Layout(**box_layout))

        self.N = IntText(description='N', style=style, layout=ipyw.Layout(width='100px'))
        self.N_box = HBox(children=(self.N, ), layout=ipyw.Layout(**box_layout))

        self.left = GridBox(children=(self.grid, self.d_box, self.N_box))
        self.children = (self.left, )

# Cell
class IndependentSamples(IndependentSamplesEstimates, Controller):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)

        self.model = model
        self.setEstimates()

        self.errBars = ErrorBars(self.model.errBars)
        self.children += (self.errBars, )


        self.grid.layout.grid_template_columns = '114px auto'
        self.m1.observe(self.observe_m1, 'value')
        self.m2.observe(self.observe_m2, 'value')
        self.sd1.observe(self.observe_sd1, 'value')
        self.sd2.observe(self.observe_sd2, 'value')
        self.ratio.observe(self.observe_ratio, 'value')
        self.N.observe(self.observe_N, 'value')

    def setEstimates(self):
        self.m1.value = self.model.m1
        self.m2.value = self.model.m2
        self.sd1.value = self.model.sd1
        self.sd2.value = self.model.sd2
        self.ratio.value = self.model.ratio
        self.d.description = str(round(self.model.d, 2))
        self.m1_label.value = 'Mean (' + self.model.label1 + ')'
        self.m2_label.value = 'Mean (' + self.model.label2 + ')'
        self.sd1_label.value = 'Standard Deviation (' + self.model.label1 + ')'
        self.sd2_label.value = 'Standard Deviation (' + self.model.label2 + ')'
        self.set_ratio_label()
        self.N.value = self.model.N

    def run(self):
        return self.model.run()

    def observe_m1(self, change):
        self.model.m1 = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setHeight(0, change['new'])
        return change['new']

    def observe_m2(self, change):
        self.model.m2 = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setHeight(1 , change['new'])
        return change['new']

    def observe_sd1(self, change):
        self.model.sd1 = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setError(0 , change['new'])
        return change['new']

    def observe_sd2(self, change):
        self.model.sd2 = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setError(1 , change['new'])
        return change['new']

    def observe_ratio(self, change):
        self.model.ratio = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.set_ratio_label()
        return change['new']

    def observe_N(self, change):
        self.model.N = change['new']
        return change['new']

    def set_ratio_label(self):
        self.ratio_label.value = 'Allocation Ratio\n (' + self.model.label1 + ': ' + str(self.model.n1) + ', ' \
            +  self.model.label2 + ': ' + str(self.model.n2) + ')'

# Cell
class PairedSamplesEstimates(Box):

    def __init__(self, **kwargs):
        lo = {'justify_content': 'space-between', 'flex_flow': 'row wrap'}
        super().__init__(layout=ipyw.Layout(**lo), **kwargs)


        ''' self.grid '''
        style={'description_width': '35px'}
        self.m1 = FloatText(step=0.01, description=r'$$M_1$$', style=style, layout=ipyw.Layout(width='100px'))
        self.m2 = FloatText(step=0.01,description=r'$$M_2$$', style=style, layout=ipyw.Layout(width='100px'))
        self.sd1 = FloatText(step=0.01,description=r'$$SD_1$$', style=style, layout=ipyw.Layout(width='100px'))
        self.sd2 = FloatText(step=0.01,description=r'$$SD_2$$', style=style, layout=ipyw.Layout(width='100px'))
        self.r = BoundedFloatText(min=-.99, max=.99, step=0.01,description=r'$$r$$', style=style, layout=ipyw.Layout(width='100px'))
        self.m1_label = AlignedLabel(value='Mean', alignment='left')
        self.m2_label = AlignedLabel(value='Mean', alignment='left')
        self.sd1_label = AlignedLabel(value='Standard Deviation', alignment='left')
        self.sd2_label = AlignedLabel(value='Standard Deviation', alignment='left')
        self.r_label = AlignedLabel(value='Correlation', alignment='left')

        self.grid = Grid(6, 2, layout=ipyw.Layout(grid_template_columns='112px auto'))
        self.grid.layout.margin = '0 0 25px 0'
        self.grid[0, 0] = AlignedLabel(value='Estimate', alignment='left')
        self.grid[0, 1] = AlignedLabel(value='Parameter', alignment='left')
        self.grid[1,0] = self.m1
        self.grid[2,0] = self.m2
        self.grid[3,0] = self.sd1
        self.grid[4,0] = self.sd2
        self.grid[5,0] = self.r
        self.grid[1, 1] = self.m1_label
        self.grid[2, 1] = self.m2_label
        self.grid[3, 1] = self.sd1_label
        self.grid[4, 1] = self.sd2_label
        self.grid[5, 1] = self.r_label

        d_label = HTMLMath(value=r'$$\frac{M_1 – M_2}{\sqrt{SD_1^2 +SD_2^2 – 2*SD_1*SD_2*r}}=d=$$', layout=ipyw.Layout(margin='5'))
        self.d = Button(disabled=True, layout=ipyw.Layout(width='65px', height='28px'))
        box_layout = {'justify_content': 'flex-end'}
        self.d_box = HBox(children=(d_label, self.d), layout=ipyw.Layout(**box_layout))

        self.N = IntText(description='N', style=style, layout=ipyw.Layout(width='100px'))
        self.N_box = HBox(children=(self.N, ), layout=ipyw.Layout(**box_layout))


        self.left = GridBox(children=(self.grid, self.d_box, self.N_box))
        self.children = (self.left, )

# Cell
class PairedSamples(PairedSamplesEstimates, Controller):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)

        self.model = model
        self.setEstimates()

        self.errBars = ErrorBars(self.model.errBars)
        self.children += (self.errBars, )

        self.grid.layout.grid_template_columns = '114px auto'
        self.m1.observe(self.observe_m1, 'value')
        self.m2.observe(self.observe_m2, 'value')
        self.sd1.observe(self.observe_sd1, 'value')
        self.sd2.observe(self.observe_sd2, 'value')
        self.r.observe(self.observe_r, 'value')
        self.N.observe(self.observe_N, 'value')

    def run(self):
        return self.model.run()

    def setEstimates(self):
        self.m1.value = self.model.m1
        self.m2.value = self.model.m2
        self.sd1.value = self.model.sd1
        self.sd2.value = self.model.sd2
        self.r.value = self.model.r
        self.d.description = str(round(self.model.d, 2))
        self.m1_label.value = 'Mean (' + self.model.label1 + ')'
        self.m2_label.value = 'Mean (' + self.model.label2 + ')'
        self.sd1_label.value = 'Standard Deviation (' + self.model.label1 + ')'
        self.sd2_label.value = 'Standard Deviation (' + self.model.label2 + ')'
        self.N.value = self.model.N

    def observe_m1(self, change):
        self.model.m1 = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setHeight(0 , change['new'])
        return change['new']

    def observe_m2(self, change):
        self.model.m2 = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setHeight(1 , change['new'])
        return change['new']

    def observe_sd1(self, change):
        self.model.sd1 = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setError(0 , change['new'])
        return change['new']

    def observe_sd2(self, change):
        self.model.sd2 = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setError(1 , change['new'])
        return change['new']

    def observe_r(self, change):
        self.model.r = change['new']
        self.d.description = str(round(self.model.d, 2))
        return change['new']

    def observe_N(self, change):
        self.model.N = change['new']
        return change['new']

# Cell
class TTestEstimates(Box):

    def __init__(self, **kwargs):
        lo = {'justify_content': 'space-between', 'flex_flow': 'row wrap'}
        super().__init__(layout=ipyw.Layout(**lo), **kwargs)

        ''' self.grid '''
        style={'description_width': '35px'}
        self.m = FloatText(description=' ', step=0.01, style=style, layout=ipyw.Layout(width='100px'))
        self.sd = FloatText(description='$$SD$$', step=0.01, style=style, layout=ipyw.Layout(width='100px'))
        self.m_label = AlignedLabel(alignment='left')
        self.sd_label = AlignedLabel(value='Standard Deviation', alignment='left')

        self.grid = Grid(3, 2, layout=ipyw.Layout(grid_template_columns='112px auto'))
        self.grid.layout.margin = '0 0 25px 0'
        self.grid[0, 0] = AlignedLabel(value='Estimate', alignment='left')
        self.grid[0, 1] = AlignedLabel(value='Parameter', alignment='left')
        self.grid[1, 0] = self.m
        self.grid[1, 1] = self.m_label
        self.grid[2, 0] = self.sd
        self.grid[2, 1] = self.sd_label

        d_label = HTMLMath(value=r'$$\frac{M}{SD}=d=$$', layout=ipyw.Layout(margin='5'))
        self.d = Button(disabled=True, layout=ipyw.Layout(width='65px', height='28px'))
        box_layout = {'justify_content': 'flex-end'}
        self.d_box = HBox(children=(d_label, self.d), layout=ipyw.Layout(**box_layout))

        self.N = IntText(description='N', style=style, layout=ipyw.Layout(width='100px'))
        self.N_box = HBox(children=(self.N, ), layout=ipyw.Layout(**box_layout))

        self.left = GridBox(children=(self.grid, self.d_box, self.N_box))
        self.children = (self.left, )

# Cell
class TTest(TTestEstimates, Controller):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)

        self.model = model

        self.errBars = ErrorBars(self.model.errBars)
        self.children += (self.errBars, )

        self.grid.layout.grid_template_columns = '114px auto'
        self.m.observe(self.observe_m, 'value')
        self.sd.observe(self.observe_sd, 'value')
        self.N.observe(self.observe_N, 'value')

        self.setEstimates()

    def run(self):
        return self.model.run()

    def setEstimates(self):
        self.m.value = self.model.m
        self.sd.value = self.model.sd
        self.d.description = str(round(self.model.d, 2))
        self.N.value = self.model.N

    def observe_m(self, change):
        self.model.m = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setHeight(0, change['new'])
        return change['new']

    def observe_sd(self, change):
        self.model.sd = change['new']
        self.d.description = str(round(self.model.d, 2))
        self.errBars.setError(0, change['new'])
        return change['new']

    def observe_N(self, change):
        self.model.N = change['new']
        return change['new']

# Cell
class SingleSample(TTest):

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

        self.m.description = r'$$M_1$$'
        self.m_label.value = r'Mean (difference from  $$M_0$$)'

# Cell
class DiffPairedSamples(TTest):

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

        self.m.description = r'$$M_D$$'
        self.m_label.value = 'Mean Difference'
        self.sd_label.value = 'Standard deviation of difference'


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_items.ipynb (unless otherwise specified).

__all__ = ['flatIndex', 'arrayIndex', 'Item', 'Collection', 'Selection', 'ItemList', 'ItemArray']

# Cell
from traitlets import HasTraits, Tuple, Unicode, Any, observe, List, Int, Float
import traitlets
import numpy as np
from traittypes import Array
from collections import namedtuple
import warnings
from IPython.display import display, Javascript

# Cell
def flatIndex(i, j, n_col):
    return i * n_col + j

def arrayIndex(index, n_col):
        row = int(index / n_col)
        col = int(index % n_col)
        return (row, col)

# Cell
class Item(HasTraits):
    '''The <code>name</code> property is used by model collections and view collections
    in order to store and retrieve models and views by name'''
    def __init__(self, name=None, **kwargs):
        super().__init__(**kwargs)
        if name:
            self.name = name
        self._index = {}

    def _repr_pretty_(self, p, cycle):
        if hasattr(self, 'name'):
            name = self.name
        else:
            name = self.__class__.__name__
        with p.group(4, str(name) +'(', ')'):
            p.breakable()
            for trait, value in self._trait_values.items():
                p.text(trait + ': ')
                p.pretty(value)
                p.text(',')
                p.breakable()

# Cell
class Collection(Item):
    '''Supports 2D and 3D arrays. 3D arrays can be accessed with tuples but are stored in
    a list.'''

    children = Tuple()

    def __init__(self, children=None, **kwargs):
        super().__init__(**kwargs)
        if children:
            self.children += tuple(children)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._children[key]
        elif isinstance(key, slice):
            return [self._children[i] for i in range(*key.indices(self._size))]
        else:
            raise KeyError('Key must be an int or slice')

    def __iter__(self):
        return self._children.__iter__()

    def get_child_index(self):
        return self._size

    @observe('children')
    def _observe_collection_children(self, change):
        self._children = ()
        self._size = 0
        for child in change['new']:
            self._add_child(child)
        return tuple(self._children)

    #def add_collection(self, collection):
    #    self._collections[id(collection)] = collection
    #    self.children += collection.children

    def _add_child(self, child):
        '''Modify child and add child to list self._children'''
        child._index[id(self)] = self.get_child_index()
        #self._indices[id(child)] = child._index[id(self)]
        self._size += 1
        self._children += (child, )

    def get_links(self, attr):
        return [(child, attr) for child in self._children]

# Cell
class Selection(Item):
    '''A collection with selectable item.'''

    children = Tuple()
    _children = Tuple()
    selected_index = Int(default_value=None, allow_none=True)
    selected_name = Unicode(default_value=None, allow_none=True)

    def __init__(self, children, **kwargs):
        super().__init__(**kwargs)
        self._size = 0
        self._children = children

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._children[key]
        elif isinstance(key, slice):
            return [self._children[i] for i in range(*key._indices(self._size))]
        elif isinstance(key, str):
            if key in self._indices:
                return self._children[self._indices[key]]
            else:
                raise KeyError('Key %s not found in children or collections' % str(key))

    def __iter__(self):
        return self._children.__iter__()

    #@validate('children')
    #TODO: children can only be changed to match selected name and index

    def add_child(self, child, name=None):
        if name:
            child.name = name
        self._children += (child, )

    @observe('_children')
    def _observe_selection_children(self, change):
        self._names = []
        self._indices = {} # name: index
        ret = tuple([self._add_child(child, i) for i, child in enumerate(change['new'])])
        if ret:
            self.selected_index = 0
            self.selected_name = ret[0].name
            self.children = (ret[0], )
        self._size = len(ret)
        return ret

    def _add_child(self, child, index):
        '''Modify child and add child to list self._children'''
        child.item_index = index
        if not hasattr(child, 'name'):
            child.name = child.__class__.__name__ + str(self._size)
        self._names.append(child.name)
        self._indices[child.name] = child.item_index
        return child

    @observe('selected_index')
    def _observe_selected_index(self, change):
        selected_index = change['new']
        if self._size:
            self.children = (self._children[selected_index], )
            self.selected_name = self._names[selected_index]
        return selected_index

    @observe('selected_name')
    def _observe_selected_name(self, change):
        selected_name = change['new']
        if self._size:
            self.selected_index = self._indices[selected_name]
        return selected_name

# Cell
class ItemList(Collection):
    '''A WidgetCollectin in which every child is the same type of Widget.
    Any child trait name listed in `trait_names` will be turned into an
    observable list, for example if 'value' is listed, a List trait 'value_list'
    will give the user access to a list of child values so that when the list
    is updated, so is the value of the child'''

    def __init__(self, children, traits=('value', ), **kwargs):
        super().__init__(**kwargs)
        self._traits = traits
        self.children = tuple(children)

    @observe('children')
    def _observe_collection_children(self, change):
        super()._observe_collection_children(change)
        if self._children and self._traits:
            for trait in self._traits:
                # initialize and observe child handler
                for child in self._children:
                    child._child_handlers[trait] = traitlets.observe(trait, type='change')
                    child._child_handlers[trait]._init_call(self._child_handler_factory(trait))
                    child.observe(child._child_handlers[trait], trait)

    def _add_child(self, child):
        child._child_handlers = {}
        super()._add_child(child)

    def set_list(self, values, trait='value'):
        if isinstance(values, np.ndarray):
            values = values.tolist()
        for i, value in enumerate(values):
            setattr(self._children[i], trait, value)

    def get_list(self, values, trait='value'):
        ret = []
        for i, child in enumerate(self._children):
            ret.append(getattr(child, trait))
        return ret

    def apply_all(self, method_name, *args, **kwargs):
        for child in self._children:
            method = getattr(child, method_name)
            method(*args, **kwargs)

    def observe_all(self, handler, names=None, type='change'):
        for child in self._children:
            if names:
                child.observe(handler, names, type)
            else:
                child.observe(handler)

    def unobserve_all(self, handler, names=None, type='change'):
        for child in self._children:
            if names:
                child.unobserve(handler, names, type)
            else:
                child.unobserve(handler)

    def observe_children(self, trait):
        for child in self._children:
            child.observe(child._child_handlers[trait], trait)

    def unobserve_children(self, trait):
        for child in self._children:
            child.unobserve(child._child_handlers[trait], trait)

    def _child_handler_factory(self, name):
        def wrapper(change):
            return self._child_handler(name, change)
        return wrapper

    def _child_handler(self, name, change):
        # add to change dict and notify change
        change['index'] = change['owner']._index[id(self)]
        change['type'] = 'child_change'
        change['list'] = self.get_list(name)
        self.notify_change(change)
        return change['new']

# Cell
class ItemArray(ItemList):
    '''A CollectionList with two dimensions. Children can be accessed
    with tuple notation similar to the numpy library. Any trait_name listed
    will cause the creation of an array of child traits, for example
    if 'value' is listed, the trait 'value_array' will become available'''
    def __init__(self, n_rows, n_columns, children=None, traits=('value',), **kwargs):
        self.n_rows = n_rows
        self.n_columns = n_columns
        super().__init__(children, traits, **kwargs)

    def get_child_index(self):
        return self._2D_index(self._size)

    def set_array(self, values, trait='value'):
        #self.unobserve_children(trait)
        if isinstance(values, np.ndarray):
            values = values.tolist()
        for i in range(self.n_rows):
            for j in range(self.n_columns):
                setattr(self._children[self._flat_index(i, j)], trait, values[i][j])
        #self.observe_children(trait)

    def get_array(self, trait='value'):
        ret = []
        for i in range(self.n_rows):
            ret.append([])
            for j in range(self.n_columns):
                ret[i].append(getattr(self._children[self._flat_index(i, j)], trait))
        return ret

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            raise IndexError('Array index must be a tuple')
        x, y = index
        if isinstance(x, int) and isinstance(y, int):
            assert x <= self.n_rows and y <= self.n_columns, "index (%i, %i) not in range" % (i, j)
            return self._children[self._flat_index(x, y)]
        else:
            ret = []
            for x in self._to_range(x, self.n_rows):
                for y in self._to_range(y, self.n_columns):
                    ret.append(self._children[self._flat_index(x, y)])
            return ret

    def _to_range(self, s, end):
        if isinstance(s, int):
            return range(s, s+1)
        else:
            if hasattr(s, 'stop') and s.stop:
                assert s.stop <= end, "slice not in range"
                end = s.stop
            return range(*s.indices(end))

    def _flat_index(self, i, j):
        return i * self.n_columns + j

    def _2D_index(self, index):
        row = int(index / self.n_columns)
        col = int(index % self.n_columns)
        return (row, col)

    def _child_handler(self, name, change):
        # add to change dict and notify change
        change['index'] = change['owner']._index[id(self)]
        change['type'] = 'child_change'
        change['array'] = self.get_array(name)
        self.notify_change(change)
        return change['new']


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_config.ipynb (unless otherwise specified).

__all__ = ['ConfigModel', 'CONFIG']

# Cell
from pathlib import Path
import os, pathlib, urllib, configparser
import rpy2.robjects
import logging

# Cell
class ConfigModel(configparser.ConfigParser):

    def __init__(self):

        ''' a custom converter we use to get a log level value from it's string representation '''
        converters={'loglevel': lambda string: getattr(logging, string)}
        super().__init__(converters=converters)

        ''' app paths'''
        self.app_dir = Path().absolute().parent

        ''' user paths '''
        self.user_dir = Path('~/.superpower_gui').expanduser()
        self.user_dir.mkdir(exist_ok=True)
        self.R_lib = self.user_dir / 'R_lib'
        self.R_lib.mkdir(exist_ok=True)
        self.user_config = self.user_dir / 'config.ini'

        ''' default config '''
        source = [str(path) for path in self.app_dir.parent.rglob('Superpower*gz')]
        source = source[0] if source else None
        defaults = {'R': {'lib_loc': self.R_lib,
                          'source': source},
                    'LOG': {'level': 'ERROR',
                            'mode': 'w',
                            'captureWarnings': True,
                            'filename': 'superpower.log'},
                    'ERRORS': {'catch_all': True}}
        self.read_dict(defaults)

        ''' read and validate user config, which overrides default '''
        if self.user_config.is_file:
            try:
                self.read(self.user_config)
                self['R']['lib_loc'] = os.path.expanduser(self['R']['lib_loc'])
                self['R']['source'] = os.path.expanduser(self['R']['source'])
                assert Path(self['R']['lib_loc']).is_dir , 'lib_loc %s is not a valid location' % self['R']['lib_loc']
                assert Path(self['R']['source']).is_file, '%s is not a valid source file' % self['R']['source']
            except Exception as e:
                pass
        self.write_user_config()

    def changeLogLevel(self):
        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        for logger in loggers:
            logger.setLevel(logging.INFO)

    def write_user_config(self):
        with open(self.user_config, 'w') as file:
            file.write('# This config file was created to match default app settings. You may want to change them.\n')
            self.write(file)

    def _repr_pretty_(self, p, cycle):
            for section in self.sections():
                for key, value in self[section].items():
                    p.text(key + ': ')
                    p.pretty(value)
                    p.breakable()

# Cell
CONFIG = ConfigModel()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/superpower.anova.ipynb (unless otherwise specified).

__all__ = ['ANOVA', 'ANCOVA', 'ANCOVAMultiway']

# Cell
from ..rpy import RFunction

# Cell
class ANOVA(RFunction):

    argNames = ['labels_factors', 'labels_levels', 'num_factors', 'num_levels', 'factorTypes', 'means',
               'SDs', 'N', 'pVal', 'full_check', 'sims', 'test_that']

    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='powerANOVAS', argNames=argNames, **kwargs)

# Cell
class ANCOVA(RFunction):
    argNames = ['labels_factors', 'labels_levels', 'num_factors', 'num_levels', 'factorTypes', 'means', 'fixedVals',
               'SDs', 'N', 'pVal', 'full_check', 'means_covariates', 'SDs_covariates', 'fixedVals',
                'labels_covariates', 'sims']

    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='powerANCOVA', argNames=argNames, **kwargs)

# Cell
class ANCOVAMultiway(RFunction):

    argNames = ['labels_factors', 'labels_levels', 'num_factors', 'num_levels', 'factorTypes', 'means', 'fixedVals',
               'SDs', 'N', 'pVal', 'full_check', 'means_covariates', 'SDs_covariates', 'fixedVals',
                'labels_covariates', 'sims']

    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='powerANCOVA_multiway', argNames=argNames, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/settings.ipynb (unless otherwise specified).

__all__ = ['LogAccordion', 'SettingsModel', 'SettingsView', 'Settings']

# Cell
from .mvc import Model, Controller
from .widget_containers import Box, Accordion, VBox, HBox
from .widgets import Dropdown, BoundedFloatText, Output, Button, BoundedIntText
from .config import CONFIG
from .log import WIDGET_HANDLER

# Cell
from traitlets import Unicode, Float, Int
import traitlets
import ipywidgets as ipyw

# Cell
class LogAccordion(Accordion):

    def __init__(self, CONFIG, WIDGET_HANDLER, **kwargs):
        super().__init__(**kwargs)
        self.config = CONFIG
        self.widget_handler = WIDGET_HANDLER
        self.widget_handler._index = {}

        # download logfile button
        html = '''
        <a download="{filename}" href="{filename}" download>
            <button class="p-Widget jupyter-widgets jupyter-button widget-button">Download Logs</button>
        </a>
        '''
        html = html.format(filename=CONFIG['LOG']['filename'])
        self.download = ipyw.HTML(html, layout=ipyw.Layout(display='inline-flex', justify_content='flex-end'))
        self.download._index = {}

        # level options
        level_options = ('DEBUG','INFO', 'WARNING','ERROR', 'CRITICAL')
        layout = ipyw.Layout(width='163px')
        style = {'description_width': 'initial'}
        self.level = Dropdown(description='Log Level',options=level_options, value=self.config['LOG']['level'], layout=layout, style=style)
        self.button = Button(description='Save', icon='save', button_style='success', layout=ipyw.Layout(width='80px'))
        #self.button.on_click(self.saveConfig)

        self.settingsBox = HBox(children=(self.level, self.button), layout=ipyw.Layout(width='100%', justify_content='flex-end'))
        self.topBox = HBox(children=(self.download, self.settingsBox))
        self.logBox = VBox(children=(self.topBox, self.widget_handler), name = 'Logs')
        self.children += (self.logBox,)

    def saveConfig(self, b):
        self.config['LOG']['level'] = str(self.level.value)
        self.config.write_user_config()
        display(Javascript('alert("Settings saved! Restart the app for changes to take effect.");'))

# Cell
class SettingsModel(Model):

    tail = Unicode()
    alpha = Float()
    sims = Int()

    def setLanding(self):
        self.tail = 'two'
        self.alpha = 0.05
        self.sims = 1

# Cell
class SettingsView(VBox):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.logAccordion = LogAccordion(CONFIG, WIDGET_HANDLER)

        self.settingsBox = Box(layout=ipyw.Layout(width='100%', flex_flow='wrap', padding='40px 0 40px 0'))
        style = {'description_width': 'initial'}
        tailLayout = ipyw.Layout(width = '163px', margin = '15px')
        alphaLayout = ipyw.Layout(width = '71px', margin = '15px')
        simsLayout = ipyw.Layout(width='150px', margin = '15px')
        tailOptions = [('2-tailed', 'two'), ('1-tailed, less', 'less'), ('1-tailed, greater', 'greater')]
        self.tail = Dropdown(options=tailOptions, description = 'Tail:', layout=tailLayout, style=style)
        self.alpha = BoundedFloatText(description='α: ', step=0.01, min = 0.0, max = 1.0, layout=alphaLayout, style=style)
        self.sims = BoundedIntText(description='# of Simulations', style=style, mix=1, layout=simsLayout)
        self.settingsBox.children += (self.tail, self.alpha, self.sims)

        self.children += (self.settingsBox, self.logAccordion)

# Cell
class Settings(SettingsView):

    def __init__(self, model=None, **kwargs):
        super().__init__(**kwargs)
        if model:
            self.introModel(model)

    def introModel(self, model):
        self.model = model
        traitlets.link((self.model, 'tail'), (self.tail, 'value'))
        traitlets.link((self.model, 'alpha'), (self.alpha, 'value'))
        traitlets.link((self.model, 'sims'), (self.sims, 'value'))


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/models.error_bars.ipynb (unless otherwise specified).

__all__ = ['ErrorBarsModel']

# Cell
from traitlets import Unicode, List, observe, Bool, Float, Int
import traitlets
from traittypes import Array
import numpy as np

# Cell
from ..mvc import Model

# Cell
class ErrorBarsModel(Model):

    '''landing'''
    nBars = Int()
    nGroups = Int()
    total = Int()

    '''estimates'''
    heights = Array(dtype=np.dtype(float))
    errors = Array(dtype=np.dtype(float))
    tops = Array(dtype=np.dtype(float))
    bots = Array(dtype=np.dtype(float))
    colors = List()
    labels = List()
    title = Unicode(allow_none=True)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def setLanding(self):
        self.nGroups = 1
        self.nBars = 1

    def setEstimates(self):
        labels = ['A', 'B', 'C', 'D', 'E', 'F']
        if self.nGroups == 1:
            self.labels = labels[:self.nBars]
        else:
            self.labels = labels[:self.nGroups]
        self.heights = np.array([100.0]*self.total)
        self.errors = np.array([10.0]*self.total)
        self.tops = np.array([105.0]*self.total)
        self.bots = np.array([95.0]*self.total)
        self.colors = ['lightblue']*self.total
        self.title = None

    def setError(self, i, error):
        self.errors[i] = error
        self.tops[i] = self.heights[i] + (error / 2)
        self.bots[i] = self.heights[i] - (error / 2)

    def setHeight(self, i, height):
        self.heights[i] = height
        self.tops[i] = height + (self.errors[i] / 2)
        self.bots[i] = height - (self.errors[i] / 2)

    def setColor(self, i, color):
        self.colors[i] = color

    @observe('nBars')
    def observe_nBars(self, change):
        self.total = change['new'] * self.nGroups

    @observe('nGroups')
    def observe_nGroups(self, change):
        self.total = self.nBars * change['new']


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/views.error_bars.ipynb (unless otherwise specified).

__all__ = ['ErrorBarsView', 'StaticErrorBars', 'ErrorBars']

# Cell
from ..widgets import Output
from ..mvc import Model
from ..items import flatIndex, arrayIndex, Item
from ..models.error_bars import ErrorBarsModel
from ..widget_containers import Box

# Cell
import matplotlib.pyplot as plt
import matplotlib.colors as mpl_colors
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
import numpy as np

# Cell
class ErrorBarsView(Output):
    def __init__(self, model, width=6, height=3, **kwargs):
        super().__init__(**kwargs)

        self.model = model

        self.bar_width = 0.8
        self.whisker_width = 0.6
        self.color_choices = {'red': mpl_colors.to_rgba('red'),
           'lightblue': mpl_colors.to_rgba('lightblue'),
           'green': mpl_colors.to_rgba('green')}

        with self:
            self.fig, self.ax = plt.subplots(figsize=[width, height])
            self.fig.canvas.toolbar_visible = False
            self.fig.canvas.header_visible = False # Hide the Figure name at the top of the figure
            self.fig.canvas.footer_visible = False
            self.fig.canvas.resizable = False
            self.fig.canvas.capture_scroll = False

        # store important values
        self.xticks = [] # locations of ticks (one for each group)
        self.xlocs = [] # one for each bar
        for i in range(self.model.nGroups):
            xlocs = [i * (self.model.nBars + 1) + j for j in range(self.model.nBars)]
            self.xlocs.extend(xlocs)
            self.xticks.append(float(sum(xlocs))/len(xlocs))

        # locations of xticks depend on the number of groups
        if self.model.nGroups == 1:
            self.ax.set_xticks(self.xlocs)
        else:
            self.ax.set_xticks(self.xticks)
        self.ax.set_xticklabels(self.model.labels)
        self.ax.yaxis.grid()
        if self.model.title:
            self.ax.set_title(self.model.title)

        # set margins (where the proportion is the space the artists take up over the xlim and ylim)
        self.ax.set_ymargin(0.4)
        if self.model.total == 1:
            self.ax.set_xmargin(0.5) # https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_xmargin.html

        self.bars = self.ax.bar(self.xlocs, self.model.heights, self.bar_width, color=self.model.colors)
        self.errbars = self.ax.errorbar(self.xlocs, self.model.heights, self.model.errors / 2,
                                        fmt='none', color='black', capsize=1)
        _, (self.top_markers, self.bot_markers) , (self.verticals,) = self.errbars
        self.top_marker_data = self.top_markers.get_ydata()
        self.bot_marker_data = self.bot_markers.get_ydata()
        self.segments = self.verticals.get_segments()

        self.lines = []
        self.tops = []
        self.bots = []
        for i in range(self.model.total):
            start = self.xlocs[i] - (self.bar_width / 2)
            end = start + self.bar_width
            line = Line2D([start, end], [self.model.heights[i]]*2, color='black', solid_capstyle='butt', linewidth=6)
            start = self.xlocs[i] - (self.whisker_width / 2)
            end = start + self.whisker_width
            top = Line2D([start, end], [self.model.tops[i]]*2, color='black', solid_capstyle='butt', linewidth=3)
            bot = Line2D([start, end], [self.model.bots[i]]*2, color='black', solid_capstyle='butt', linewidth=3)
            self.ax.add_artist(line)
            self.ax.add_artist(top)
            self.ax.add_artist(bot)
            self.lines.append(line)
            self.tops.append(top)
            self.bots.append(bot)


# Cell
class StaticErrorBars(ErrorBarsView):

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

    def setColor(self, i , color):
        self.bars[i].set_facecolor(self.color_choices[color])
        self.model.colors[i] = color
        change = {'new': color, 'index': i, 'name': 'colors', 'type': 'child_change'}
        self.notify_change(change)

    def _setHeight(self, i, height):
        self.model.setHeight(i, height)
        self.bars[i].set_height(height)
        self.lines[i].set_ydata([height, height])
        self._setHeightError(i, height, self.model.errors[i])
        change = {'new': height, 'index': i, 'name': 'heights', 'type': 'child_change'}
        self.notify_change(change)

    def setHeight(self, i, height):
        self._setHeight(i, height)
        self.redrawCanvas()

    def setError(self, i, error):
        self._setError(i, error)
        self.redrawCanvas()

    def _setError(self, i, error):
        self.model.setError(i, error)
        self._setHeightError(i, self.model.heights[i], error)
        change = {'new': error, 'index': i, 'name': 'errors', 'type': 'child_change'}
        self.notify_change(change)

    def _setHeightError(self, i, height, error):

        # whiskers
        self.tops[i].set_ydata(self.model.tops[i])
        self.bots[i].set_ydata(self.model.bots[i])

        # markers (which are covered up by whiskers but are important)
        self.top_marker_data[i] = self.model.tops[i]
        self.bot_marker_data[i] = self.model.bots[i]
        self.top_markers.set_ydata(self.top_marker_data) # update view components
        self.bot_markers.set_ydata(self.bot_marker_data)

        # vertical lines
        self.segments[i][0][1] = self.model.tops[i]
        self.segments[i][1][1] = self.model.bots[i]
        self.verticals.set_segments(self.segments)

    def redrawCanvas(self):
        self.ax.relim()
        self.ax.autoscale()
        self.fig.canvas.flush_events()
        self.fig.canvas.draw()

# Cell
class ErrorBars(StaticErrorBars):
    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

        self.cidpress = self.fig.canvas.mpl_connect(
           'button_press_event', self.onPress)
        self.cidrelease = self.fig.canvas.mpl_connect(
            'button_release_event', self.onRelease)
        self.cidmotion = None
        self.picked = None
        self.event_start = None

    def onPress(self, event):
            if event.inaxes != self.ax: return
            for i in range(self.model.total):
                self._onPress(i, event)

    def _onPress(self, i, event):
        ''' returns true for anything that needs motion'''
        x0 = self.bars[i].get_x()
        x1 = x0 + self.bar_width
        if event.xdata > x0 and event.xdata < x1:
            if self.lines[i].contains(event)[0]:
                self.event_start = event.ydata
                self.height_start = self.model.heights[i]
                self.picked = i
                self.cidmotion = self.fig.canvas.mpl_connect('motion_notify_event', self.onLineMotion)
            elif self.tops[i].contains(event)[0]:
                self.event_start = event.ydata
                self.error_start = self.model.errors[i]
                self.picked = i
                self.cidmotion = self.fig.canvas.mpl_connect('motion_notify_event', self.onTopMotion)
            elif self.bots[i].contains(event)[0]:
                self.event_start = event.ydata
                self.error_start = self.model.errors[i]
                self.picked = i
                self.cidmotion = self.fig.canvas.mpl_connect('motion_notify_event', self.onBotMotion)
            elif event.ydata > 0 and event.ydata < self.bars[i].get_height():
                current = self.bars[i].get_facecolor()
                if current == self.color_choices['lightblue']:
                    self.setColor(i, 'red')
                elif current == self.color_choices['red']:
                    self.setColor(i, 'green')
                else:
                    self.setColor(i, 'lightblue')

    def onLineMotion(self, event):
        if self.picked is not None:
            dy = event.ydata - self.event_start
            height = self.height_start + dy
            self._setHeight(self.picked, height)

    def onTopMotion(self, event):
        if event.inaxes != self.ax: return
        if self.picked is not None and self.event_start is not None:
            print(self.picked, event)
            dy = event.ydata - self.event_start
            error = self.error_start + dy * 2
            self._setError(self.picked, error)

    def onBotMotion(self, event):
        if event.inaxes != self.ax: return
        if self.picked is not None:
            print(self.picked, event)
            dy = self.event_start - event.ydata
            error = self.error_start + dy * 2
            self._setError(self.picked, error)

    def onRelease(self, event):
        if event.inaxes != self.ax: return
        if self.picked is not None:
            self.cidmotion = self.fig.canvas.mpl_disconnect(self.cidmotion)
            self.picked = None
            self.redrawCanvas()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_rpy.ipynb (unless otherwise specified).

__all__ = ['warnError', 'rPrint', 'R_CONSOLE', 'import_rpy2', 'pkg_name', 'utils', 'lib_loc', 'libPaths', 'SUPERPOWER',
           'types', 'DTYPES', 'AS_DTYPE', 'Result', 'rpy2py_intvector', 'rpy2py_floatvector', 'rpy2py_strvector',
           'rpy2py_boolvector', 'rpy2py_listsexpvector', 'converter', 'rpy2py_langsexpvector',
           'rpy2py_pairlistsexpvector', 'py2rpy_list', 'py2rpy_list', 'py2rpy_none', 'RFunction', 'MatrixMixin']

# Cell
import time, os, datetime
import pandas as pd
import numpy as np
from IPython.display import display, Javascript, HTML
from collections import UserDict
from pathlib import Path

# Cell
from .log import getLogger, WIDGET_HANDLER
from .config import CONFIG

# Cell
import rpy2
from rpy2 import rinterface, robjects, situation
rpy2.rinterface.initr()

from rpy2.robjects import packages, numpy2ri, pandas2ri, conversion
from rpy2.rinterface_lib.embedded import RRuntimeError

# Cell
R_CONSOLE = getLogger('r.console')
def warnError(msg):
    R_CONSOLE.error(msg)

def rPrint(msg):
    R_CONSOLE.info(msg)

rpy2.rinterface_lib.callbacks.consolewrite_warnerror = warnError
rpy2.rinterface_lib.callbacks.consolewrite_print = rPrint

# Cell
def import_rpy2():
    global rpy2
    return rpy2

# Cell
pkg_name = 'Superpower'
utils = rpy2.robjects.packages.importr('utils')
lib_loc = CONFIG['R']['lib_loc']
libPaths = list(rpy2.robjects.r['.libPaths']())
libPaths.insert(0, CONFIG['R']['lib_loc'])
SUPERPOWER = None

for lib_path in libPaths:
    try:
        R_CONSOLE.info('Attempting Superpower import from %s...' % lib_path)
        SUPERPOWER = rpy2.robjects.packages.importr(pkg_name, lib_loc)
        if SUPERPOWER:
            R_CONSOLE.info('Import successful!')
            break
    except:
        pass
if not SUPERPOWER:
    R_CONSOLE.exception('Import failed.')
    R_CONSOLE.info('Using source file %s' % lib_loc)
    R_CONSOLE.info('Attempting install to %s' % lib_loc)
    try:
        kwargs = {}
        kwargs['pkgs'] = CONFIG['R']['source']
        kwargs['repos'] = rpy2.rinterface.NULL
        kwargs['type'] = 'source'
        kwargs['lib'] = lib_loc
        utils.install_packages(**kwargs)
        SUPERPOWER = rpy2.robjects.packages.importr(pkg_name, lib_loc)
    except Exception as e:
        R_CONSOLE.exception('Install failed with message: %s' % str(e))
        raise Exception('Fatal. Superpower could not be imported or installed.')
    if SUPERPOWER:
        R_CONSOLE.info('Install successful!')
    else:
        raise Exception('Install failed.')

# Cell
types = (('int', rpy2.rinterface.IntSexpVector([])), ('bool', rpy2.rinterface.BoolSexpVector([])), ('string', rpy2.rinterface.StrSexpVector([])), ('float', rpy2.rinterface.FloatSexpVector([])))
DTYPES = {}
AS_DTYPE = {}
for py_type, r_vector in types:
    np_array = np.asarray(r_vector)
    DTYPES[py_type] = np_array.dtype
    AS_DTYPE[py_type] = getattr(np, str(np_array.dtype))

# Cell
class Result(UserDict):

    def _repr_pretty_(self, p, cycle):
        with p.group(4, self.data['name'] +'([', '])'):
            p.breakable()
            for key, value in self.data.items():
                if key != 'name':
                    p.text(key  + ': ')
                    if isinstance(value, dict):
                        with p.group(4, '{', '}'):
                            for k, v in value.items():
                                p.text(str(k) + ': ')
                                p.pretty(v)
                                p.text(', ')
                                p.breakable()
                    elif isinstance(value, pd.DataFrame):
                        p.pretty('See above data frame.')
                        display(HTML(value.to_html()))
                    else:
                        p.pretty(value)
                        p.text(',')
                    p.breakable()

# Cell
converter = conversion.Converter('superpower converter')

@converter.rpy2py.register(rinterface.IntSexpVector)
def rpy2py_intvector(obj):
    if 'factor' in obj.rclass or len(obj) != 1:
        return robjects.pandas2ri.rpy2py(obj)
    else:
        return obj[0]

@converter.rpy2py.register(rinterface.FloatSexpVector)
def rpy2py_floatvector(obj):
    res = robjects.pandas2ri.rpy2py_floatvector(obj)
    if len(res) == 1:
        return obj[0]
    else:
        return res

@converter.rpy2py.register(rinterface.StrSexpVector)
def rpy2py_strvector(obj):
    if len(obj) == 1:
        return obj[0]
    else:
        return [s for s in obj]

@converter.rpy2py.register(rinterface.BoolSexpVector)
def rpy2py_boolvector(obj):
    if len(obj) == 1:
        return obj[0]
    else:
        return [s for s in obj]

@converter.rpy2py.register(rinterface.ListSexpVector)
def rpy2py_listsexpvector(obj):
    # recursive base case
    if not isinstance(obj, rinterface.ListSexpVector):
        if isinstance(obj, rinterface.NULLType):
            return None
        else:
            return converter.rpy2py(obj)

    if 'data.frame' in obj.rclass: # if data frame
        res = pd.DataFrame(robjects.ListVector(obj)) #robjects.pandas2ri.rpy2py_dataframe(obj)
    elif isinstance(obj.names, rinterface.NULLType): # if list isn't named
        return [rpy2py_listsexpvector(value) for value in obj]
    else: # if named list
        res = {}
        for key, value in zip(obj.names, obj):
            res[key] = rpy2py_listsexpvector(value)
    return res

# this kills creation of ListVectors
#@converter.rpy2py.register(rinterface.NULLType)
#def rpy2py_null(obj):
#    return None

# Cell
@converter.rpy2py.register(rinterface.LangSexpVector)
def rpy2py_langsexpvector(obj):
    ''' This is for when there is an r default like "c('one', 'two')" '''
    obj = [rpy2py_strvector(item) for item in obj if isinstance(item, rinterface.StrSexpVector)]
    return obj

@converter.rpy2py.register(rinterface.PairlistSexpVector)
def rpy2py_pairlistsexpvector(obj):
    ''' This one is important for getting default arguments with the formals function'''
    ret = {}
    for item in obj:
        if item in converter.rpy2py.registry.keys():
            ret.update(rpy2py_listsexpvector(item))
        if item in robjects.default_converter.rpy2py.registry.keys():
            ret.update(rpy2py_listsexpvector(item))
    return ret

# Cell
@converter.py2rpy.register(list)
def py2rpy_list(obj):
    if len(obj) == 1:
        with conversion.localconverter(robjects.default_converter + robjects.numpy2ri.converter + robjects.pandas2ri.converter + converter) as cv:
            return cv.py2rpy(obj[0])
    if len(obj) > 1:
        first = obj[0]
        if all(isinstance(x, type(first)) for x in obj):
            if isinstance(first, bool):
                return rinterface.BoolSexpVector(obj)
            elif isinstance(first, str):
                return rinterface.StrSexpVector(obj)
            elif isinstance(first, int):
                return rinterface.IntSexpVector(obj)
            elif isinstance(first, float):
                return rinterface.FloatSexpVector(obj)
    return robjects.default_converter.py2rpy(obj)

# Cell
@converter.py2rpy.register(bool)
def py2rpy_list(obj):
    return rinterface.BoolSexpVector([obj])

@converter.py2rpy.register(type(None))
def py2rpy_none(obj):
    return robjects.NULL

# Cell
class RFunction():
    ''' The class used to create a model for any R function, and only one R function.
    This class abstracts away all rpy2 calls to make interaction with the SUPERPOWER library
    easy.
    '''

    def __init__(self, name, argNames=None):
        self.name = name
        self.function = getattr(SUPERPOWER, name)
        # converter defined in global scope above
        self.converter = robjects.default_converter + robjects.numpy2ri.converter + robjects.pandas2ri.converter + converter
        self.defaults = self.getDefaultKwargs()
        self.logger = getLogger(__name__)

    def setArgNames(self, argNames=None):
        '''Sets the names of all accepted arguments. Overrides previous.'''
        if argNames is None:
            self.getArgNames(self)
        else:
            if all(name in self.argNames for name in argNames):
                self.argNames = argNames
            else:
                raise Exception('All argNames must be valid arguments for the given R function')

    def setKwargs(self, **kwargs):
        """Adds kwargs if and only if the key is is a valid function argument.
        """
        self.kwargs = {k:v for (k, v) in kwargs.items() if k in self.argNames}

    def getArgNames(self):
        '''Gets the names of all accepted arguments'''
        formals = rpy2.robjects.r['formals']
        default = formals(self.function)
        return default.names

    def getDefaultKwargs(self):
        '''This method gets a dictionary of default arguments provided by the R function'''
        with conversion.localconverter(self.converter) as cv:
            defaults = self.function.formals()
            ret = {}
            return self.null_to_none(defaults)

    def py2rpy(self):
        self.rargs = {}
        self.rargsString = ''
        with conversion.localconverter(self.converter) as cv:
            for key, value in self.kwargs.items():
                rpy = cv.py2rpy(value)
                self.rargs[key] = rpy
                if hasattr(rpy, 'r_repr'):
                    self.rargsString += '\t%s: %s\n' % (key, rpy.r_repr())
                else:
                    self.rargsString += '\t%s: %s\n' % (key, str(rpy))
            return self.rargsString

    def run(self, alert=True, **kwargs):
        '''This method runs the R function associated with the model. The run will be
        timed and the runtime and results are added to the results dictionary. You should
        not have to modify this function to run your custom Model
        '''
        if kwargs:
            self.setKwargs(**kwargs)
        try:
            self.py2rpy()
            ret = {}
            startTime = time.time()
            ret['name'] = self.name
            ret['kwargs'] = self.kwargs
            ret['rargs'] = self.rargs
            self.logger.info('Running %s with R arguments\n%s' % (self.name, self.rargsString))
            with conversion.localconverter(self.converter):
                ret['results'] = self.null_to_none(self.function(**self.rargs))
            ret['runtime'] = (time.time() - startTime)
            ret['id'] = int(round(time.time() * 1000))
            ret['date'] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            self.logger.info('ID: %i, Date: %s, Runtime: %s' % (ret['id'], str(ret['date']), str(ret['runtime'])))
            return Result(ret)
        except Exception as e:
            if alert:
                estring = str(type(e).__name__) + ': '
                estring += str(e).replace('\n', ' ').replace('\n', '').replace('"','\\"')
                display(Javascript('alert("' + estring + '");'))
                if isinstance(type(e), RRuntimeError):
                    robjects.r('traceback(max.lines=1)')
                else:
                    self.logger.error(e, exc_info=True)
                return None
            else:
                raise e

    # we want the converter to do this but it causes problems
    def null_to_none(self, dictionary):
        ret = {}
        for key, value in dictionary.items():
            if isinstance(value, rinterface.NULLType):
                ret[key] = None
            else:
                ret[key] = value
        return ret

    def attr_dict(self):
        keys = ['name', 'result', 'packageName', 'runtime', 'id', 'kwargs', 'defaults']
        return {key: value for key, value in self.__dict__.items() if key in keys}

    def _repr_pretty_(self, p, cycle):
        with p.group(4, str(self.name) +'([', '])'):
            p.breakable()
            for key, value in self.attr_dict().items():
                if key != 'name':
                    p.text(key  + ': ')
                    if isinstance(value, dict):
                        with p.group(4, '{', '}'):
                            for k, v in value.items():
                                p.text(k + ': ')
                                p.pretty(v)
                                p.text(',')
                                p.breakable()
                    else:
                        p.pretty(value)
                        p.text(',')
                        p.breakable()
                    p.breakable()

# Cell
class MatrixMixin():

    def matrixRFunction(self, **kwargs):
        func = robjects.r('matrix')
        with robjects.conversion.localconverter(robjects.default_converter + robjects.numpy2ri.converter):
            return func(**kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/main.ipynb (unless otherwise specified).

__all__ = ['Node', 'MainView', 'Main']

# Cell
from IPython.display import display, Javascript, HTML
import json
import traitlets
from traitlets import HasTraits, Int, Unicode, observe
import ipywidgets as ipyw
from IPython import get_ipython

# Cell
from .widgets import IntSlider, Button, Dropdown, FloatText, RadioButtons, Output
from .widget_containers import Tab, SelectionVBox, VBox, Box, Accordion
from .views.tabs import LandingTab, EstimatesTab, ResultsTab, SettingsTab
from .models.chisq import *
from .views.chisq import *
from .models.corr import *
from .views.corr import *
from .models.anova import *
from .views.anova import *
from .models.ttest import *
from .views.ttest import *
from .results import ResultsModel, Results
from .settings import SettingsModel, Settings
from superpower_gui import log
from .config import CONFIG

# Cell
class Node:

    decisionWidgets = ipyw.VBox()
    landingWidgets = ipyw.Box()
    view = ipyw.VBox(children = (decisionWidgets, landingWidgets))

    ''' builds a decision tree where all children are either Models or Widgets'''
    def __init__(self, key, label, widget=None, model=None, landing=None, Estimates=None):
        if key == 'head':
            self.decisionWidgets.children = ()
            self.landingWidgets.children = ()
        self.children = {}
        self.selected = None
        self.widget = widget
        if self.widget:
            self.widget.observe(self.observeChange, names='value')
            self.decisionWidgets.children += (self.widget,)
        self.key = key
        self.label = label
        self.model = model
        self.landing = landing
        self.Estimates = Estimates

    def __getattr__(self, attr):
        return self.children[attr]

    def addModel(self, key, label, model, Landing, Estimates):
        if Landing:
            landing = Landing(model)
        else:
            landing = Output() # a "blank" widget
        self.children[key] = Node(key, label, model=model, landing=landing, Estimates=Estimates)
        self.selected = key

    def addWidget(self, key, label, widget):
        self.children[key] = Node(key, label, widget)

    def select(self, key):
        if key in self.children.keys():
            self.selected = key

    def getModel(self):
        child = self.children[self.selected]
        if child.model:
            return child
        else:
            return child.getModel()

    def observeChange(self, change):
        self.selected = change['new']
        self.showWidgets()
        self.landingWidgets.children = (self.getModel().landing, )

    def showWidgets(self):
        for key, child in self.children.items():
            if key == self.selected:
                if child.widget:
                    child.widget.show()
                child.showWidgets()
            else:
                if child.widget:
                    child.widget.hide()
                child.hideWidgets()

    def hideWidgets(self):
        for key, child in self.children.items():
            if child.widget:
                child.widget.hide()
            child.hideWidgets()

    def addOptions(self):
        for key, child in self.children.items():
            if self.widget:
                self.widget.options = [(child.label, key) for key, child in self.children.items()]
            child.addOptions()

    def getWidgetList(self, wlist):
        if self.widget:
            self.widget.options = [(child.label, key) for key, child in self.children.items()]
            wlist += [self.widget]
        for key, child in self.children.items():
            wlist = child.getWidgetList(wlist)
        return wlist

    def init(self):
        self.addOptions()
        self.showWidgets()
        self.landingWidgets.children = (self.getModel().landing, )

# Cell
class MainView(ipyw.Tab):

    TAB_TITLES = ('Landing', 'Estimates', 'Results', 'Settings')

    def __init__(self, **kwargs):
        super().__init__(layout=ipyw.Layout(height='100%'), **kwargs)

        # landing tab
        self.tree = Node('head', 'Model Class', Dropdown(description='Model Class', layout=ipyw.Layout(max_width='300px', description_width='80px')))
        self.tree.addWidget('corr', 'Correlation', Dropdown(description='Coefficient', layout=ipyw.Layout(max_width='350px',description_width='100px')))
        self.tree.addWidget('ttest', 'T-Test', Dropdown(description='Model'))
        self.tree.ttest.addWidget('paired', 'Paired Sample T-Test', RadioButtons())

        self.proceedButton = Button(description='Proceed', button_style='success')
        self.landingTab = ipyw.VBox(children = (self.tree.view, self.proceedButton, ))

        # estimates tab
        self.estimatesBox = ipyw.Box()
        self.calculateButton = Button(description='Calculate', button_style='success')
        self.estimatesTab = ipyw.VBox(children = (self.estimatesBox, self.calculateButton))

        # results tab
        self.results = Results()
        self.repopLandingButton = Button(description='Repopulate Model', layout=ipyw.Layout(width='180px'))
        self.repopEstimatesButton = Button(description='Repopulate Estimates', layout=ipyw.Layout(width='200px'))
        buttonBox = ipyw.HBox(children=(self.repopLandingButton, self.repopEstimatesButton))
        self.resultsTab = ipyw.VBox(children=(self.results, buttonBox))

        # settings tab
        self.settingsTab = Settings()

        # main
        self.children = (self.landingTab, self.estimatesTab, self.resultsTab, self.settingsTab)
        for i, title in enumerate(self.TAB_TITLES):
            self.set_title(i, title)

# Cell
class Main(MainView):

    def __init__(self):
        super().__init__()

        # models
        self.models = {}
        self.models['chisq'] = ChiSqModel()
        self.models['anova'] = ANOVAModel()
        self.models['tetra'] = TetraModel()
        self.models['cramerv'] = CramerVModel()
        self.models['pears'] = PearsonModel()
        self.models['spear'] = SpearmansRhoModel()
        self.models['point'] = PBiModel()
        self.models['single'] = SingleSampleModel()
        self.models['ind'] = IndependentSamplesModel()
        self.models['pair'] = PairedSamplesModel()
        self.models['diff'] = DiffPairedSamplesModel()
        self.models['results'] = ResultsModel()
        self.models['settings'] = SettingsModel()

        # views
        self.tree.addModel('chisq', 'Chi Squared', self.models['chisq'], ChiSqLanding, ChiSq)
        self.tree.addModel('anova', 'Analysis of Variance', self.models['anova'], ANOVALanding, ANOVAEstimates)
        self.tree.corr.addModel('tetra', 'Tetrachoric', self.models['tetra'], TetraLanding, Tetra)
        self.tree.corr.addModel('cramerv', "Cramer's V", self.models['cramerv'], CramerVLanding, CramerV)
        self.tree.corr.addModel('pears', "Pearson's Correlation Coefficient", self.models['pears'], PearsonLanding, Pearson)
        self.tree.corr.addModel('spear', "Spearman's Rho", self.models['spear'], SpearmansRhoLanding, SpearmansRho)
        self.tree.corr.addModel('point', "Point Biserial", self.models['point'], PBiLanding, PBi)
        self.tree.ttest.addModel('single', 'One Sample T-Test', self.models['single'], None , SingleSample)
        self.tree.ttest.addModel('ind', 'Independent Samples T-Test', self.models['ind'], TTestLanding, IndependentSamples)
        self.tree.ttest.paired.addModel('diff', 'From Mean difference and SD of difference', self.models['diff'], None, DiffPairedSamples)
        self.tree.ttest.paired.addModel('pair', 'From Means and Standard Deviations', self.models['pair'], None, PairedSamples)
        self.tree.init()

        self.results.introModel(self.models['results'])
        self.settingsTab.introModel(self.models['settings'])

        # observe
        self.proceedButton.on_click(self.proceed)
        self.calculateButton.on_click(self.calculate)
        self.repopLandingButton.on_click(self.repopulateLanding)
        self.repopEstimatesButton.on_click(self.repopulateEstimates)

    def proceed(self, button):
        self.model = self.tree.getModel()
        self.model.model.setEstimates()
        self.estimatesBox.children = (self.model.Estimates(self.model.model), )
        self.selected_index = 1

    def calculate(self, button):
        self.updateModelWithSettings()
        result = self.model.model.run()
        if result:
            self.results.addResult(result)
            self.selected_index = 2

    def updateModelWithSettings(self):
        if self.model.key in ('chisq','tetra', 'cramerv'):
            self.model.model.pVal = self.models['settings'].alpha
            self.model.model.tail = self.models['settings'].tail
        else:
            self.model.model.setTraits(**self.models['settings'].getTraits())

    def repopulateLanding(self, button):
        self.selected_index = 0

    def repopulateEstimates(self, button):
        self.selected_index = 1


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_log.ipynb (unless otherwise specified).

__all__ = ['DispatchingFormatter', 'FORMATTER', 'ROOT_LOGGER', 'FILE_HANDLER', 'WidgetHandler', 'WIDGET_HANDLER',
           'getLogger', 'logger', 'loadCSS']

# Cell
import ipywidgets as ipyw
import traitlets
import warnings
import os, pathlib, urllib
from IPython.display import HTML, Javascript, display
from pathlib import Path
import logging
from .config import CONFIG

# Cell
class DispatchingFormatter:
    """Dispatch formatter for logger and it's sub logger."""
    def __init__(self):
        self.formatters = {
            'root': logging.Formatter('%(message)s'),
            'r.console': logging.Formatter('[R.%(levelname)s] %(message)s'),
            #'js.output': logging.Formatter('[JS.OUTPUT] %(message)s'),
            #'js': logging.Formatter('[JS.%(levelname)s] %(message)s'),
            'py.warnings': logging.Formatter('[PY.WARNINGS] %(message)s'),
            '__main__': logging.Formatter('[PY.%(levelname)s] %(message)s'),
            'default': logging.Formatter('[PY.%(levelname)s](%(name)s:%(lineno)d)  %(message)s')
        }

    def format(self, record):
        formatter = self.formatters.get(record.name, self.formatters['default'])
        return formatter.format(record)

# Cell
FORMATTER = DispatchingFormatter()

# Cell
ROOT_LOGGER = logging.getLogger('')
ROOT_LOGGER.setLevel(CONFIG['LOG']['level'])

# Cell
FILE_HANDLER = logging.FileHandler(CONFIG['LOG']['filename'],
                               CONFIG['LOG']['mode'])
FILE_HANDLER.setFormatter(FORMATTER)
logging.getLogger('').addHandler(FILE_HANDLER)

# Cell
class WidgetHandler(ipyw.Textarea, logging.Handler):

    def __init__(self, config):
        super().__init__(layout={'width': '100%', 'height': '250px'}, disabled=True)
        self.config = config

    def emit(self, record):
        self.value += str(self.format(record)) + '\n'

# Cell
WIDGET_HANDLER = WidgetHandler(CONFIG)
WIDGET_HANDLER.setFormatter(FORMATTER)
logging.getLogger('').addHandler(WIDGET_HANDLER)

# Cell
def getLogger(name=''):
    return logging.getLogger(name)

# Cell
logger = getLogger()
logger.info('Welcome to Superpower!')

# Cell
logging.captureWarnings(CONFIG['LOG']['captureWarnings'])

# Cell
def loadCSS():
    css_path = CONFIG.app_dir / 'assets/css'
    logger = logging.getLogger(__name__)
    logger.info('Loading css files...')
    for file in list(css_path.rglob('*.css')):
        with open(file, 'r') as f:
            try:
                filestr = '<style>\n' + f.read() + '\n</style>'
                display(HTML(filestr))
                logger.info('\t%s' % file)
            except:
                logger.exception('Error loading css file %s' % file)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/views.anova.ipynb (unless otherwise specified).

__all__ = ['ANOVALandingView', 'ANOVALanding', 'ANOVAUnit', 'ANOVAEstimatesView', 'ANOVAEstimates']

# Cell
from ..models.error_bars import ErrorBarsModel
from .error_bars import ErrorBars
from ..items import ItemList, ItemArray
from ..widget_containers import Box, VBox, Tab, HBox
from ..widgets import Dropdown, Text, BoundedIntText, BoundedFloatText, Checkbox, AlignedLabel,Textarea, Button, \
RadioButtons, GridBox, Grid, FloatText, IntText

# Cell
import ipywidgets as ipyw
from traitlets import observe
import traitlets
from IPython.display import display, Javascript

# Cell
class ANOVALandingView(Box):

    def __init__(self, **kwargs):
        layout = {'flex_flow': 'row wrap', 'justify_content': 'space-between'}
        super().__init__(layout = ipyw.Layout(**layout), **kwargs)

        boxLayout = {'display': 'flex', 'flex_flow': 'column nowrap', 'align_items': 'flex-start'}

        style={'description_width': 'initial'}

        factorsOptions = [i for i in range(1, 6)]
        self.num_factors = Dropdown(value=None, description='Factors', options=factorsOptions, layout=ipyw.Layout(width='100px', margin='10px'), style=style)

        self.leftBox = Box(children=(self.num_factors,),layout=ipyw.Layout(**boxLayout))
        self.levelsBox = Box(layout=ipyw.Layout(**boxLayout, margin='10px'))

        self.factorLabels = ItemList([Text(value=None, description='Factor', layout=ipyw.Layout(width='100px'), style=style) for i in range(5)])
        self.numLevels = ItemList([Dropdown(value=None, description='Levels',style=style,options=[i for i in range(2,7)], layout=ipyw.Layout(width='95px')) for i in range(5)])
        self.levelLabels = ItemArray(5, 6, children=[Text(value=None, layout=ipyw.Layout(width='62px')) for i in range(30)])
        self.factor_types = ItemList([RadioButtons(value=None, options=(('Between', 'b'), ('Within','w')),layout=ipyw.Layout(width='100px')) for i in range(5)])
        self.factorBoxes = []

        for i in range(0, 5):
            self.numLevels[i].observe(self.numLevelsChanged, names='value')
            self.factor_types[i].observe(self.factor_typesChanged, names='value')
            box = Box(children=(self.factorLabels[i], self.numLevels[i]) + tuple(self.levelLabels[i, :]) +  (self.factor_types[i],), layout=ipyw.Layout(display='align_items'))
            self.factorBoxes.append(box)
            self.levelsBox.children += (box, )

        self.r = BoundedIntText(value=None, description=r'$r_{within-subject factors}$', min=0, max=10, style=style, layout=ipyw.Layout(width='170px', margin='10px'))
        self.num_covariates = BoundedIntText(value=None, description='Covariates', min=0, max=10, style=style, layout=ipyw.Layout(width='135px', margin='10px'))
        self.powerCovariates = Checkbox(value=None, description='Power Covariates', style=style, margin='10px')
        self.leftBox.children += (self.levelsBox, self.r, self.num_covariates, self.powerCovariates)

        label = AlignedLabel(value=r'Adjusted $\beta$')
        self.covariatesBox = Box(children=(label,), layout=ipyw.Layout(**boxLayout))
        covariateLabels = [Text(value=None, description='Covariate', layout=ipyw.Layout(width='135px'), style=style) for i in range(0, 10)]
        self.covariateLabels = ItemList(covariateLabels)
        covariates = [BoundedFloatText(value=None, step=0.01, layout=ipyw.Layout(width='60px'), style=style) for i in range(0, 10)]
        self.covariates = ItemList(covariates)
        self.covariateBoxes = []

        #self.covariateBoxes = []
        for i in range(0, 10):
            box = Box(children=(self.covariateLabels[i], self.covariates[i]), layout=ipyw.Layout(display='align_items'))
            self.covariateBoxes.append(box)
            self.covariatesBox.children += (box, )

        self.children = (self.leftBox, self.covariatesBox)

        ''' observes '''
        self.num_factors.observe(self.num_factorsChanged, names='value')
        self.powerCovariates.observe(self.powerCovariatesChanged, names='value')
        self.num_covariates.observe(self.num_covariatesChanged, names='value')

    def num_factorsChanged(self, change):
        for i in range(1, change['new']):
            self.factorBoxes[i].visible()
        for i in range(change['new'], 5):
            self.factorBoxes[i].invisible()

    def powerCovariatesChanged(self, change):
        if change['new'] == False:
            display(Javascript('alert("If you do not power your covariates, you must enter adjusted means for each factor.");'))

    def num_covariatesChanged(self, change):
        if change['new'] == 0:
            self.powerCovariates.invisible()
            self.covariatesBox.hide()
        else:
            self.powerCovariates.visible()
            self.covariatesBox.show()
            for i in range(0, change['new']):
                self.covariateBoxes[i].visible()
            for i in range(change['new'], 10):
                self.covariateBoxes[i].invisible()

    def numLevelsChanged(self, change):
        for i in range(2, change['new']):
            self.levelLabels[change['index'],i].visible()
        for i in range(change['new'], 6):
            self.levelLabels[change['index'],i].invisible()

    def factor_typesChanged(self, change):
        if change['new'] == 'w':
            self.r.visible()
        else:
            if 'w' not in self.factor_types.get_list('value')[:self.num_factors.value]:
                self.r.invisible()

# Cell
class ANOVALanding(ANOVALandingView):

    def __init__(self, model, **kwargs):
        super().__init__()

        self.model = model

        self.numLevels.observe(self.observeNumLevels, names='value',type='child_change')
        self.levelLabels.observe(self.observeLevelLabels, names='value',type='child_change')
        self.factorLabels.observe(self.observeFactorLabels, names='value',type='child_change')
        self.covariateLabels.observe(self.observeCovariateLabels, names='value',type='child_change')
        self.factor_types.observe(self.observeFactorTypes, names='value',type='child_change')
        self.covariates.observe(self.observeCovariates, names='value', type='child_change')

        traitlets.link((self.model, 'num_factors'), (self.num_factors, 'value'))
        traitlets.link((self.model, 'r'), (self.r, 'value'))
        traitlets.link((self.model, 'num_covariates'), (self.num_covariates, 'value'))
        traitlets.link((self.model, 'powerCovariates'), (self.powerCovariates, 'value'))

        self.setLanding()

        # force observe method even if the value hasn't changed
        self.num_covariates._notify_trait('value', None, self.model.num_covariates)
        self.powerCovariates._notify_trait('value', None, self.model.powerCovariates)

    def setLanding(self):
        self.levelLabels.set_array(self.model.levelLabels)
        self.factor_types.set_list(self.model.factor_types)
        self.covariateLabels.set_list(self.model.covariateLabels)
        self.factorLabels.set_list(self.model.factorLabels)
        self.covariates.set_list(self.model.w_covariates)
        self.numLevels.set_list(self.model.numLevels)

        self.num_factors.value = self.model.num_factors
        self.r.value = self.model.r
        self.num_covariates.value = self.model.num_covariates
        self.powerCovariates.value = self.model.powerCovariates


    def observeCovariates(self, change):
        self.model.changeCovariates(change['index'], change['new'])

    def observeNumLevels(self, change):
        self.model.changeNumLevels(change['index'], change['new'])

    def observeLevelLabels(self, change):
        i, j = change['index']
        self.model.changeLevelLabel(i, j, change['new'])

    def observeFactorLabels(self, change):
        self.model.changeFactorLabel(change['index'], change['new'])

    def observeFactorTypes(self, change):
        self.model.changeFactorType(change['index'], change['new'])

    def observeCovariateLabels(self, change):
        self.model.changeCovariateLabel(change['index'], change['new'])

# Cell
class ANOVAUnit(Box):

    def __init__(self, model, style='single', **kwargs):
        super().__init__(**kwargs)
        assert style =='single' or style =='multi', 'Style choices are "single" or "multi"'
        self.model = model

        self.M = ItemList([FloatText(step=0.1,layout=ipyw.Layout(width='50px')) for i in range(self.model.total)])
        self.SD = ItemList([FloatText(step=0.1,layout=ipyw.Layout(width='50px')) for i in range(self.model.total)])
        self.N = ItemList([IntText(layout=ipyw.Layout(width='50px')) for i in range(self.model.total)])

        self.grids = []
        if self.model.nGroups == 1:
            for i in range(self.model.nBars):
                grid = Grid(3, 2, layout=ipyw.Layout(justify_items='flex-end'), name=self.model.labels[i])
                grid.add_widget_list([self.M[i]], 0, 1)
                grid.add_widget_list([self.SD[i]], 1, 1)
                grid.add_widget_list([self.N[i]], 2, 1)
                grid[0, 0] = AlignedLabel(value='M')
                grid[1, 0] = AlignedLabel(value = 'SD')
                grid[2, 0] = AlignedLabel(value = 'N')
                self.grids.append(grid)

        else:
            for i in range(self.model.nGroups):
                grid = Grid(3, self.model.nBars + 1, layout=ipyw.Layout(justify_items='flex-end'), name=self.model.labels[i])
                start = i * self.model.nBars
                end = start + self.model.nBars
                grid.add_widget_list(self.M[start:end], 0, 1)
                grid.add_widget_list(self.SD[start:end], 1, 1)
                grid.add_widget_list(self.N[start:end], 2, 1)
                grid[0, 0] = AlignedLabel(value='M')
                grid[1, 0] = AlignedLabel(value = 'SD')
                grid[2, 0] = AlignedLabel(value = 'N')
                self.grids.append(grid)

        self.errorBars = ErrorBars(self.model, layout=ipyw.Layout(display = 'inline-flex'))

        self.layout.flex_flow = 'column nowrap'
        self.layout.width= 'min-content'
        # tabbed or no?
        if style == 'multi' or self.model.total > 6:
            multi={'width': '100%', 'padding': '0 66px 0 80px'}
            self.tab = Tab(children = self.grids, layout=ipyw.Layout(**multi))
            self.tabBox = Box(children=(self.tab,))
            self.children = (self.errorBars, self.tabBox)
            if style == 'multi':
                self.tabBox.layout.padding='38px 0 38px 0'
                self.layout.flex_flow = 'row nowrap'
                self.layout.display = 'flex'
            else:
                self.tab

        else:
            layout = {'justify_content': 'space-around', 'padding': '0 66px 0 42px'}
            self.box = HBox(children=self.grids, layout=ipyw.Layout(**layout))
            #self.layout.flex_flow = 'col nowrap'
            self.children = (self.errorBars, self.box)


         # observe grid
        self.M.observe(self.observeM, names='value', type='child_change')
        self.SD.observe(self.observeSD, names='value', type='child_change')
        self.N.observe(self.observeN, names='value', type='child_change')
        self.errorBars.observe(self.observeError, names='errors', type='child_change')
        self.errorBars.observe(self.observeHeight, names='heights', type='child_change')

        self.setEstimates()

    def flatIndex(self, i, j):
        ret = i * self.model.nBars + j
        return ret

    def gridIndex(self, i):
        grid = i % self.model.nGroups
        index = grid + 1
        return grid, index

    def setEstimates(self):
        for i in range(self.model.nGroups):
            for j in range(self.model.nBars):
                flat = self.flatIndex(i,j)
                self.M[flat].value = self.model.heights[flat]
                self.SD[flat].value = self.model.errors[flat]
                self.N[flat].value = self.model.Ns[flat]

    def observeHeight(self, change):
        grid, index = self.gridIndex(change['index'])
        self.grids[grid][0, index] = change['new']

    def observeM(self, change):
        self.errorBars.setHeight(change['index'], round(change['new'], 1))

    def observeSD(self, change):
        self.errorBars.setError(change['index'], round(change['new'], 1))

    def observeN(self, change):
        self.model.Ns[change['index']] = change['new']

    def observeHeight(self, change):
        self.M[change['index']].value = round(change['new'], 1)

    def observeError(self, change):
        self.SD[change['index']].value = round(change['new'], 1)

# Cell
class ANOVAEstimatesView(VBox):

    def __init__(self, nBars, nGroups, nGraphs, **kwargs):
        lo = {'justify_content': 'space-between', 'flex_flow': 'row wrap', 'width': '100%'}
        super().__init__(layout = ipyw.Layout(**lo), **kwargs)
        style={'description_width': '20px'}

        ''' left '''
        self.left = VBox(name='left', layout=ipyw.Layout(min_width='250px', width='30%'))
        self.textarea = Textarea(layout=ipyw.Layout(width="94%", height="150px"), disabled=True)
        self.effectButton = Button(description='Show effect sizes', button_style='success',
            layout=ipyw.Layout(min_width='220px', max_width='220px',margin = '2px 2px 2px 2px'))
        self.left.children = (self.textarea, self.effectButton)

        ''' right '''
        self.right = VBox(name='right', layout=ipyw.Layout(width='auto'))

        self.children =  (self.right, )#(self.left, self.right) <--- uncomment to include effect size

# Cell
class ANOVAEstimates(ANOVAEstimatesView):

    def __init__(self, model, **kwargs):
        super().__init__(nBars=model.nBars, nGroups=model.nGroups, nGraphs=model.nUnits, **kwargs)
        self.model = model

        if model.nUnits == 1:
            style = 'single'
        else:
            style = 'multi'

        self.units = []
        for i in range(self.model.nUnits):
            self.units.append(ANOVAUnit(self.model.units[i], style=style))
            self.right.children += (self.units[i], )

    def run(self):
        return self.model.run()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/models.anova.ipynb (unless otherwise specified).

__all__ = ['ANOVALandingModel', 'ANOVAUnitModel', 'ANOVAModel']

# Cell
from ..mvc import Model
from .error_bars import ErrorBarsModel
from ..superpower.anova import ANOVA

# Cell
from traitlets import Int, List, Float, Bool, observe
import copy
import numpy as np
from traittypes import Array

# Cell
class ANOVALandingModel(Model):

    ''' var names from powerANCOVA_multiway.R'''
    num_factors = Int()
    num_levels = List() # Number of groups being tested. Should be a vector if more than one factor
    labels_factors = List()
    labels_levels = List()
    num_covariates = Int(1) # that 1 is there for a reason
    labels_covariates = List(allow_none=True)
    factorTypes = List()

    factor_types = List()
    w_covariates = Array(dtype=np.dtype(float))
    covariates = Array(allow_none=True, dtype=np.dtype(float))
    numLevels = List()
    factorLabels = List()
    levelLabels = List()
    covariateLabels = List()
    r = Float()
    powerCovariates = Bool()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def setLanding(self):
        with self.hold_trait_notifications():
            self.numLevels = [2] * 5
            self.num_levels = [2] * 5
            self.factorLabels = ['A', 'B', 'C', 'D', 'E']
            self.levelLabels = [[str(i) for i in range(1,7)] for j in range(5)]
            self.labels_levels = copy.deepcopy(self.levelLabels)
            self.factor_types = ['b']*5
            self.covariateLabels = ['CV'+str(i) for i in range(1, 11)]
            self.w_covariates = np.full(10, 0.0)
            self.covariates = np.full(10, 0.0)
            self.r = 0.0
            self.powerCovariates = True
            self.num_covariates = 0
        for i in range(5):
            self.changeNumLevels(i, 2)
        self.num_factors = 1


    @observe('num_factors')
    def observe_num_factors(self, change):
        self.labels_factors = self.factorLabels[:change['new']]
        self.factorTypes = self.factor_types[:change['new']]
        self.num_levels = self.numLevels[:change['new']]
        self.labels_levels = [self.levelLabels[i][:self.num_levels[i]] for i in range(change['new'])]

    @observe('num_covariates')
    def changeNumCovariates(self, change):
        if change['new'] == 0:
            self.labels_covariates = None
            self.covariates = None
        else:
            self.labels_covariates = self.covariateLabels[:change['new']]
            self.covariates = self.w_covariates[:change['new']]

    def changeCovariates(self, i, value):
        self.w_covariates[i] = value
        if self.num_covariates > i:
            self.covariates[i] = value

    def changeNumLevels(self, i, num):
        self.numLevels[i] = num
        if self.num_factors > i:
            self.num_levels[i] = num
            self.labels_levels[i] = self.levelLabels[i][:num]

    def changeLevelLabel(self, i, j, value):
        self.levelLabels[i][j] = value
        if self.num_factors > i:
            if self.num_levels[i] > j:
                self.labels_levels[i][j] = value

    def changeCovariateLabel(self, i, value):
        self.covariateLabels[i] = value
        if self.num_covariates > i:
            self.labels_covariates[i] = value

    def changeFactorLabel(self, i, value):
        self.factorLabels[i] = value
        if self.num_factors > i:
            self.labels_factors[i] = value

    def changeFactorType(self, i , value):
        self.factor_types[i] = value
        if self.num_factors > i:
            self.factorTypes[i] = value

# Cell
class ANOVAUnitModel(ErrorBarsModel):

    Ns = List()

    def flatIndex(self, i, j):
        return i * self.nBars + j

    def setEstimates(self):
        super().setEstimates()
        for j in range(self.nGroups):
            for k in range(self.nBars):
                self.setHeight(self.flatIndex(j, k), 1 + 0.1 * k)
                self.setError(self.flatIndex(j, k), 1)
                self.Ns.append(10)

    def getMeans(self):
        ret = []
        for j in range(self.nGroups):
            start = self.nBars*j
            ret.append(self.heights[start:start+self.nBars])
        return ret

    def getSDs(self):
        ret = []
        for j in range(self.nGroups):
            start = self.nBars*j
            ret.append(self.errors[start:start+self.nBars])
        return ret

    def getNs(self):
        ret = []
        for j in range(self.nGroups):
            start = self.nBars*j
            ret.append(self.Ns[start:start+self.nBars])
        return ret

# Cell
class ANOVAModel(ANOVALandingModel):

    nBars = Int()
    nGroups = Int()
    nUnits = Int()
    labels = List()
    means = List()
    SDs = List()
    N = List()
    sims = Int()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.functions = {}
        self.functions['anova'] = ANOVA()

    def run(self, **kwargs):
        if self.nUnits > 1:
            self.means = [unit.getMeans() for unit in self.units]
            self.SDs = [unit.getSDs() for unit in self.units]
            self.N = [unit.getNs() for unit in self.units]
        elif self.nUnits == 1:
            self.means = self.units[0].getMeans()
            self.SDs = self.units[0].getSDs()
            self.N = self.units[0].getNs()
        return self.functions['anova'].run(**self.getTraits())

    def flatIndex(self, i, j):
        return i * self.nBars + j

    def setEstimates(self):

        self.labels = []
        for i, f in enumerate(self.labels_factors):
            self.labels.append([f + ':' + l for l in self.labels_levels[i]])

        self.sims = 1
        self.nBars = self.num_levels[-1]
        self.nGroups = 1
        self.units = []
        if self.num_factors >= 2:
            self.nGroups = self.num_levels[-2]
        elif self.num_factors < 3:
            self.units = [ANOVAUnitModel(nBars=self.nBars, nGroups=self.nGroups)]
            self.units[0].setEstimates()
            self.units[0].title = None
            self.units[0].labels = self.labels[0]

        if self.num_factors == 3:
            self.nUnits = len(self.labels[0])
            for a in self.labels[0]:
                unit = ANOVAUnitModel(nBars=self.nBars, nGroups=self.nGroups)
                unit.setEstimates()
                unit.title = a
                unit.labels = self.labels[1]
                self.units.append(unit)
        elif self.num_factors == 4:
            self.nUnits = len(self.labels[0]) * len(self.labels[1])
            for a in self.labels[0]:
                for b in self.labels[1]:
                    unit = ANOVAUnitModel(nBars=self.nBars, nGroups=self.nGroups)
                    unit.setEstimates()
                    unit.title = a + ' X ' + b
                    unit.labels = self.labels[2]
                    self.units.append(unit)
        elif self.num_factors == 5:
            self.nUnits = len(self.labels[0]) * len(self.labels[1]) * len(self.labels[2])
            for a in self.labels[0]:
                for b in self.labels[1]:
                    for c in self.labels[2]:
                        unit = ANOVAUnitModel(nBars=self.nBars, nGroups=self.nGroups)
                        unit.setEstimates()
                        unit.title = a + ' X ' + b + ' X ' + c
                        unit.labels = self.labels[3]
                        self.units.append(unit)

        self.nUnits = len(self.units)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/superpower.ttest.ipynb (unless otherwise specified).

__all__ = ['TTest']

# Cell
from ..rpy import RFunction

# Cell
class TTest(RFunction):

    argNames = ['N', 'n1', 'n2', 'ratio', 'sims', 'meanDifference',
               'means1', 'sd1', 'means2', 'sd2', 'var_equal', 'type',
               'corr', 'altHypothesis', 'variances', 'pVal']

    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='powerTtest', argNames=argNames, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/controller.ipynb (unless otherwise specified).

__all__ = ['Selection', 'Controller']

# Cell
from IPython.display import display, Javascript, HTML
import json
import traitlets
from traitlets import HasTraits, Int, Unicode, observe
import ipywidgets as ipyw
from IPython import get_ipython

# Cell
from .widgets import IntSlider, Button, Dropdown, FloatText
from .widget_containers import Tab, SelectionVBox, VBox, Box, Accordion
from .views.tabs import LandingTab, EstimatesTab, ResultsTab, SettingsTab
from .models.chisq import *
from .views.chisq import *
from .models.corr import *
from .views.corr import *
from .models.anova import *
from .views.anova import *
from .results import ResultsModel, Results
from .settings import SettingsModel, Settings
from superpower_gui import log
from .config import CONFIG

# Cell
from collections import namedtuple
Selection = namedtuple('Selection', ['label', 'LandingModel', 'Model', 'Landing', 'Estimates'])

# Cell
class Controller(traitlets.HasTraits):

    modelName = Unicode()
    corrModelName = Unicode()

    def __init__(self):

        ''' log handlers'''
        self.logger = log.getLogger(__name__)
        self.logger.info('Autosave: off')
        display(Javascript('Jupyter.notebook.set_autosave_interval(0);'))
        self.logger.info('Output area scroll: off')
        display(Javascript('IPython.OutputArea.prototype._should_scroll = function(lines){return false;}'))
        self.logger.info('Set container width to 100%')
        display(HTML("<style>.container { width:100% !important; }</style>"))
        log.loadCSS()

        ''' this wonderful piece of magic lets us catch global ipython exceptions
        that are usually handled by calling showtraceback that pretty prints the
        traceback beneath the cell in which the error occured. This is important,
        because ipywidget and traitlets have a lot of callbacks, and these can't be
        caught with a simple try/catch block.
        https://github.com/ipython/ipython/blob/5da6784b690b085bd9cde88e08ab4e2def9f65ff/IPython/core/interactiveshell.py
        and search for dummy_handler for original example'''


        if CONFIG['ERRORS'].getboolean('catch_all'):
            python_exception_types = (Exception, StopIteration, SystemExit, ArithmeticError, OverflowError,
                     FloatingPointError, ZeroDivisionError, AssertionError, AttributeError, EOFError,
                     ImportError, KeyboardInterrupt, LookupError, IndexError, KeyError, NameError,
                     UnboundLocalError, EnvironmentError, IOError, OSError, SyntaxError, IndentationError,
                     SystemError, SystemExit, TypeError, ValueError, RuntimeError, NotImplementedError)
            get_ipython().set_custom_exc(python_exception_types, self.catch_all)

        ''' create views and models '''
        self.create_global_widgets()
        self.define_box_selections()
        self.initialize_mvcs()
        self.view = self.tab_based_view()
        self.view.add_class('main_tab')
        self.view.layout.height = '100%'
        self.add_observations()

        ''' initialize '''
        self.corrDropdown.value = 'cramerv'
        self.dropdown.value = 'chisq'

    def catch_all(self, ip, etype, value, tb, tb_offset=None):
        if hasattr(etype, '__name__') and value:
            self.logger.error(etype.__name__ + ': ' + str(value))
            display(Javascript('alert("' + etype.__name__ + ': ' + str(value) + '\\n\\nIf a problem occurs, restart the app. For farther information, check the log accordion in the settings tab.")'))
        else:
            self.logger.error('An error has occurred.')
            display(Javascript('alert("An error has occured. Check the log accordion in the settings tab for more details, or restart the app.")'))
        if tb:
            self.logger.error(str(tb))

    def _ipython_display_(self):
        display(self.view)
        self.logger.info('Cell height expanded.')
        js = """
        var main_cell = $('.main_tab').parent().parent();
        var height = $('#site').height()
        main_cell.height(height - 100);
        """
        display(Javascript(js))

    def create_global_widgets(self):

        ''' landing '''
        self.corrDropdown = Dropdown(description='Coefficient: ',
            layout=ipyw.Layout(max_width='350px',description_width='100px'))
        self.corrSelection = SelectionVBox()
        self.corrBox = VBox(children=(self.corrDropdown, self.corrSelection))

        self.landingSelection = SelectionVBox()
        self.dropdownOptions = [('Correlation', 'corr')] # populated in self.initialize_view_and_models
        self.dropdown = Dropdown( options=self.dropdownOptions, description='Model Class: ',
                layout=ipyw.Layout(max_width='300px', description_width='80px'))

        self.okayButton = Button(description='Proceed', button_style='success')

        ''' estimates '''
        self.estimatesBox = Box()
        self.runButton = Button(description='Calculate', button_style='success')
        self.restartButton = Button(description='Restart', layout=ipyw.Layout(width='90px'))

        ''' results '''
        style = {'description_width': 'initial'}
        self.repopulateLanding = Button(description='Repopulate Model Tab', layout=ipyw.Layout(width='180px'))
        self.repopulateEstimates = Button(description='Repopulate Estimates Tab', layout=ipyw.Layout(width='200px'))

    def define_box_selections(self):

        self.options = {}
        self.options['chisq'] = Selection('Chi Squared', ChiSqLandingModel, ChiSqModel, ChiSqLanding, ChiSq)
        self.options['anova'] = Selection('Analysis of Variance', ANOVALandingModel, ANOVAModel, ANOVALanding, ANOVAEstimates)

        self.corrOptions = {}
        self.corrOptions['tetra'] = Selection('Tetrachoric', TetraLandingModel, TetraModel, TetraLanding, Tetra)
        self.corrOptions['cramerv'] = Selection("Cramer's V", CramerVLandingModel, CramerVModel, CramerVLanding, CramerV)
        self.corrOptions['pears'] = Selection("Pearson's Correlation Coefficient", CorrLandingModel, PearsonModel, PearsonLanding, Pearson)
        self.corrOptions['spear'] = Selection("Spearman's Rho", CorrLandingModel, SpearmansRhoModel, SpearmansRhoLanding, SpearmansRho)
        self.corrOptions['point'] = Selection("Point Biserial", CorrLandingModel, PBiModel, PBiLanding, PBi)

    def initialize_single_mvc(self, options, selection, dropdown):
        for name, option in options.items():
            label, LandingModel, Model, Landing, Estimates = option
            landingModel = LandingModel()
            self.landingModels[name] = landingModel
            self.Model[name] = Model
            dropdown.options += ((label, name), )
            selection.add_child(Landing(landingModel), name)
            self.Estimates[name] = Estimates

    def initialize_mvcs(self):

        ''' selectable boxes in landing '''
        self.Model = {}
        self.Estimates = {}
        self.landingModels = {}
        self.initialize_single_mvc(self.options, self.landingSelection, self.dropdown)

        self.corrLandingModels = {}
        self.initialize_single_mvc(self.corrOptions, self.corrSelection, self.corrDropdown)
        self.landingSelection.add_child(self.corrBox, 'corr')

        '''other mvcs'''
        self.resultsModel = ResultsModel()
        self.settingsModel = SettingsModel()

        self.resultsView = Results(self.resultsModel)

    def tab_based_view(self):
        landingTab = LandingTab(self.dropdown, self.landingSelection, self.okayButton, name='Model')
        estimatesTab = EstimatesTab(self.estimatesBox, self.runButton, name='Estimates')
        resultsTab = ResultsTab(self.restartButton, self.repopulateLanding, self.repopulateEstimates, self.resultsView, name='Results')
        settingsTab = Settings(self.settingsModel, name='Settings')
        return Tab(children=(landingTab, estimatesTab, resultsTab, settingsTab))

    def add_observations(self):
        self.okayButton.on_click(self.on_okayButton_click)
        self.runButton.on_click(self.on_runButton_click)
        self.restartButton.on_click(self.on_restartButton_click)
        self.repopulateLanding.on_click(self.on_repopulateLanding_click)
        self.repopulateEstimates.on_click(self.on_repopulateEstimates_click)

        self.dropdown.observe(self.on_dropdown_change, names='value')
        self.corrDropdown.observe(self.on_corrDropdown_change, names='value')

    def on_dropdown_change(self, change):
        if change['new'] == 'corr':
            self.modelName = self.corrDropdown.value
            self.landingSelection.selected_name = 'corr'
        else:
            self.modelName = change['new']
            self.landingSelection.selected_name = self.modelName

    def on_corrDropdown_change(self, change):
        self.corrSelection.selected_name = change['new']
        self.modelName = change['new']

    ''' on button click'''
    def on_okayButton_click(self, button):
        model = self.Model[self.modelName](**self.landingModels[self.modelName].getTraits())
        model.setEstimates()
        self.estimates = self.Estimates[self.modelName](model, name=self.modelName)
        self.estimatesBox.children = (self.estimates, )
        self.view.selected_name = 'Estimates'

    def on_runButton_click(self, button):
        self.update_model_with_settings()
        result = self.estimates.run()
        if result:
            self.resultsView.addResult(result)
            self.view.selected_name='Results'

    def on_restartButton_click(self, button):
        if self.modelName == 'corr':
            self.corrSelection[self.corrModelName].setLanding()
        else:
            self.landingSelection[self.modelName].setLanding()
        self.view.selected_name= 'Model'

    def on_repopulateLanding_click(self, button):
        self.view.selected_name = 'Model'

    def on_repopulateEstimates_click(self, button):
        self.view.selected_name = 'Estimates'

    def update_model_with_settings(self):
        if self.modelName in ('chisq','tetra', 'cramerv'):
            self.estimates.model.pVal = self.settingsModel.alpha
            self.estimates.model.tail = self.settingsModel.tail
        else:
            self.estimates.model.setTraits(**self.settingsModel.getTraits())


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/models.ttest.ipynb (unless otherwise specified).

__all__ = ['TTestModel', 'SingleSampleModel', 'DiffPairedSamplesModel', 'PairedSamplesModel', 'IndependentSamplesModel']

# Cell
from traitlets import Float, Int, observe, Bool, Unicode
import math

# Cell
from ..mvc import Model
from .error_bars import ErrorBarsModel
from ..superpower.ttest import TTest

# Cell
class TTestModel(Model):

    m = Float()
    sd = Float()
    d = Float()
    N = Int()

    def __init__(self):
        super().__init__()
        self.function = TTest()

    def setLanding(self):
        self.errBars = ErrorBarsModel(nBars=1, nGroups=1)
        self.errBars.setEstimates()
        self.errBars.labels=['']

    def setEstimates(self):
        with self.hold_trait_notifications():
            self.m = 0.2
            self.sd = 1
            self.N = 200


    @observe('m')
    def observe_mean(self, change):
        self.d = change['new'] / self.sd
        self.errBars.setHeight(0, change['new'])
        return change['new']

    @observe('sd')
    def observe_sd(self, change):
        self.d = self.m / change['new']
        self.errBars.setError(0, change['new'])
        return change['new']

# Cell
class SingleSampleModel(TTestModel):

    def run(self):
        kwargs = self.getTraits()
        kwargs['type'] = 'singleSample'
        kwargs['means1'] = kwargs.pop('m')
        kwargs['sd1'] = kwargs.pop('sd')
        return self.function.run(**kwargs)

class DiffPairedSamplesModel(TTestModel):

    def run(self):
        #res = function.run(meanDifference = .2, sd1=1, N=200, type='pairedSamples')
        kwargs = self.getTraits()
        kwargs['type'] = 'pairedSamples'
        kwargs['sd1'] = kwargs.pop('sd')
        kwargs['meanDifference'] = kwargs.pop('m')
        return self.function.run(**kwargs)

# Cell
class PairedSamplesModel(Model):

    label1 = Unicode()
    label2 = Unicode()

    m1 = Float()
    m2 = Float()
    sd1 = Float()
    sd2 = Float()
    d = Float()
    N = Int()
    r = Float()

    def __init__(self):
        super().__init__()
        self.function = TTest()

    def run(self):
        kwargs = self.getTraits()
        kwargs['corr'] = kwargs.pop('r')
        kwargs['type'] = 'pairedSamples'
        kwargs['means1'] = kwargs.pop('m1')
        kwargs['means2'] = kwargs.pop('m2')
        return self.function.run(**kwargs)

    def setLanding(self):
        self.errBars = ErrorBarsModel(nBars=2,nGroups=1)
        self.errBars.setEstimates()
        self.label1 = 'A'
        self.label2 = 'B'

    def setEstimates(self):
        self.errBars.setEstimates()
        with self.hold_trait_notifications():
            self.m1 = 1
            self.m2 = 1.2
            self.sd1 = 1
            self.sd2 = 1
            self.N = 200
            self.r = .8

    @observe('label1')
    def observe_label1(self, change):
        self.errBars.labels[0] = change['new']
        return change['new']

    @observe('label2')
    def observe_label2(self, change):
        self.errBars.labels[1] = change['new']
        return change['new']

    @observe('m1')
    def observe_m1(self, change):
        self.errBars.setHeight(0, change['new'])
        self.calc_d(change['new'], self.m2, self.r, self.sd1, self.sd2)
        return change['new']

    @observe('m2')
    def observe_m2(self, change):
        self.errBars.setHeight(1, change['new'])
        self.calc_d(self.m1, change['new'], self.r, self.sd1, self.sd2)
        return change['new']

    @observe('r')
    def observe_ratio(self, change):
        assert change['new'] < 1, 'Ratio must be less than 1.'
        self.calc_d(self.m1, self.m2, change['new'], self.sd1, self.sd2)
        return change['new']

    @observe('sd1')
    def observe_sd1(self, change):
        self.errBars.setError(0, change['new'])
        self.calc_d(self.m1, self.m2, self.r, change['new'], self.sd2)
        return change['new']

    @observe('sd2')
    def observe_sd2(self, change):
        self.errBars.setError(1, change['new'])
        self.calc_d(self.m1, self.m2, self.r, self.sd1, change['new'])
        return change['new']

    def calc_d(self, m1, m2, r, sd1, sd2):
        self.d = round((m1 - m2) / math.sqrt(sd1**2 + sd2**2 - 2 * sd1 * sd2 * r), 2)

# Cell
class IndependentSamplesModel(Model):

    label1 = Unicode()
    label2 = Unicode()

    m1 = Float()
    m2 = Float()
    sd1 = Float()
    sd2 = Float()
    d = Float()
    N = Int()
    ratio = Float()
    n1 = Int()
    n2 = Int()

    def __init__(self):
        super().__init__()
        self.function = TTest()

    def run(self):
        kwargs = self.getTraits()
        kwargs['type'] = 'indSamples'
        kwargs['means1'] = kwargs.pop('m1')
        kwargs['means2'] = kwargs.pop('m2')
        kwargs.pop('n1')
        kwargs.pop('n2')
        return self.function.run(**kwargs)

    def setLanding(self):
        self.errBars = ErrorBarsModel(nBars=2,nGroups=1)
        self.errBars.setEstimates()
        self.label1 = 'A'
        self.label2 = 'B'


    def setEstimates(self):
        with self.hold_trait_notifications():
            self.m1 = 1
            self.m2 = 1.2
            self.sd1 = 1
            self.sd2 = 1
            self.ratio = 1
            self.N = 200
            self.n1


    @observe('label1')
    def observe_label1(self, change):
        self.errBars.labels[0] = change['new']
        return change['new']

    @observe('label2')
    def observe_label2(self, change):
        self.errBars.labels[1] = change['new']
        return change['new']

    @observe('m1')
    def observe_m1(self, change):
        self.errBars.setHeight(0, change['new'])
        self.calc_d(change['new'], self.m2, self.ratio, self.sd1, self.sd2)
        return change['new']

    @observe('m2')
    def observe_m2(self, change):
        self.errBars.setHeight(1, change['new'])
        self.calc_d(self.m1, change['new'], self.ratio, self.sd1, self.sd2)
        return change['new']

    @observe('ratio')
    def observe_ratio(self, change):
        self.calc_d(self.m1, self.m2, change['new'], self.sd1, self.sd2)
        n2 = self.N / (change['new'] + 1)
        self.n1 = int(n2 * change['new'])
        self.n2 = int(n2)
        return change['new']

    @observe('sd1')
    def observe_sd1(self, change):
        self.errBars.setError(0, change['new'])
        self.calc_d(self.m1, self.m2, self.ratio, change['new'], self.sd2)
        return change['new']

    @observe('sd2')
    def observe_sd2(self, change):
        self.errBars.setError(1, change['new'])
        self.calc_d(self.m1, self.m2, self.ratio, self.sd1, change['new'])
        return change['new']

    def calc_d(self, m1, m2, ratio, sd1, sd2):
        self.d = round((m1 - m2) / math.sqrt((ratio * sd1**2 + sd2**2) / (ratio + 1)), 2)